/*
**    TP ETI3 CPE Lyon
**    Copyright (C) 2009  Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef _MC_NAVIGATOR_TOOL_HPP_
#define _MC_NAVIGATOR_TOOL_HPP_

#include <MC_v3d.hpp>
#include <MC_trackball.hpp>


namespace mesh_conv
{

    /// Container class for the viewer, does not need to be weel written
    class MC_navigator_tool
    {

        public:

        MC_navigator_tool();

        //*********************************************//
        //ORIENTATION
        //*********************************************//

        // orientation of camera 1
        MC_trackball trackball_cam1;
        // translation of the camera 1
        MC_v3d tr_cam1;
        // distance to camera 1
        double dist_cam1;


        // orientation of light 1
        MC_trackball trackball_light1;
        // translation of the light 1
        MC_v3d tr_light1;
        // distance to light 1
        double dist_light1;

        // Matrix light1
        MC_matrix current_light1() const;
        //  get the matrix parameters for light1
        /** \return position,normal,right,up*/
        MC_v3d_vector axis_light1() const;


        // Matrix camera1
        MC_matrix current_cam1() const;
        //  get the matrix parameters
        /** \return position,normal,right,up*/
        MC_v3d_vector axis_cam1() const;

        // move camera trackball (forward=q*(0,0,1))
        void go_forward_trackball_cam1(const double& d_L);
        // move camera trackball (right=q*(-1,0,0))
        void go_right_trackball_cam1(const double& d_L);
        // move camera trackball (up=q*(0,-1,0))
        void go_up_trackball_cam1(const double& d_L);

        // save cam1
        std::ostream& write_cam1(std::ostream& stream) const;
        // save cam1 to file
        const MC_navigator_tool& write_cam1_file(std::string& filename) const;
        //load cam1
        MC_navigator_tool& read_cam1(std::istream& param);
        //load cam1 file
        MC_navigator_tool& read_cam1_file(const std::string& filename);

        // save light1
        std::ostream& write_light1(std::ostream& stream) const;
        // save light1 to file
        const MC_navigator_tool& write_light1_file(std::string& filename) const;
        //load light1
        MC_navigator_tool& read_light1(std::istream& param);
        //load light1 file
        MC_navigator_tool& read_light1_file(const std::string& filename);


        //*********************************************//
        //TIME
        //*********************************************//

        /// return current_time-old_time;
        double delta_time() {return current_time-old_time;}
        /// update the fps and return fps
        double update_fps();
        /// increase the frame drawn and return the number
        int increase_frame_number() {++frame;return frame;}
        unsigned int current_time,old_time,frame;
        double current_fps;

        //*********************************************//
        //Mouse record
        //*********************************************//

        int x_previous,y_previous;

        //*********************************************//
        //Screen Size
        //*********************************************//

        int x_screen_size_1,y_screen_size_1;

        //*********************************************//
        //Projection matrix
        //*********************************************//

        double fov,znear,zfar;

        // left/top/znear/zfar/aspect
        /* for use: glFrustum(left, -left, top, -top, znear, zfar);
           with
             double aspect=static_cast<double>(navigator.x_screen_size_1)/static_cast<double>(navigator.y_screen_size_1);
             double fov=45.0*PI/180.0;
             double near=0.1;
             double top=tan(fov*0.5)*near;
             double bottom=-top;
             double left=aspect*bottom;
             double right=-left;
           */
        MC_double_vector projection_param() const;

        /* glFrustum camera
          http://www.opengl.org/sdk/docs/man/xhtml/glFrustum.xml
          */
        MC_matrix projection_matrix() const;

        //*********************************************//
        //Screen tool record
        //*********************************************//

        double size_pointer;


        //*********************************************//
        //Picking ray
        //*********************************************//

        /** \brief give the (x,y) coordinate in screen space and return the center and the ray pointed in world space
            \return center,ray
            */
        MC_v3d_vector ray_world_space_cam1(const int& x_screen,const int& y_screen) const;

        //*********************************************//
        //mouse
        //*********************************************//
        bool left_button;
        bool right_button;

        private:
    };
}

#endif
