
/*
**    TP ETI3 CPE Lyon
**    Copyright (C) 2009  Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef   	MC_TRACKBALL_HPP_
# define   	MC_TRACKBALL_HPP_



#include <iostream>


#include <MC_quaternion.hpp>

namespace mesh_conv
{

    /// A container class to deal with a trackball
    class MC_trackball
    {
    public:

        //*********************************************//
        //*********************************************//
        // CONSTRUCTORS
        //*********************************************//
        //*********************************************//

        /** \empty initial constructor */
        MC_trackball();

        /** \brief classical destructor */
        ~MC_trackball();

        //*********************************************//
        //*********************************************//
        // Get quaternion
        //*********************************************//
        //*********************************************//

        /** \brief get the current quaternion */
        const MC_quaternion& quaternion() const;
        /** \brief get the current quaternion */
        MC_quaternion& quaternion();

        /** \brief get the quaternion to apply */
        const MC_quaternion& d_quaternion() const;
        /** \brief get the quaternion to apply */
        MC_quaternion& d_quaternion();

        //*********************************************//
        //*********************************************//
        // Manipulation
        //*********************************************//
        //*********************************************//

        /** \brief return the distance to disc for the 3D coordinate */
        double project_to_disc(const double& x,const double& y) const;

        /** \brief given a set of 2d coords difference, return the corresponding quaternion */
        MC_trackball& set_2d_coords(const double& x0,const double& y0,const double& x1,const double& y1);

        /** \brief multiply d_q to current_q */
        MC_trackball& apply_rotation();

        /** \brief Set d_q to 0 */
        MC_trackball& no_motion();


        /** \brief internal parameter of the disc radius */
        double disc_radius;

    private:



        /** \brief the current quaternion */
        MC_quaternion current_q;
        /** \brief the quaternion to apply */
        MC_quaternion d_q;
    };

}

#endif
