/*
**    TP ETI3 CPE Lyon
**    Copyright (C) 2009 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <algorithm>
#include <fstream>
#include <cmath>

#include <MC_string_helper.hpp>

namespace mesh_conv
{
  

  // helper to tokenize
  std::vector<std::string> MC_string_tokenizer::tokenize(const std::string& str,const std::string& delimiters)
  {
    std::vector<std::string> tokens;
    std::string::size_type delimPos = 0, tokenPos = 0, pos = 0;
    
    if(str.length()<1)  return tokens;
    while(1){
      delimPos = str.find_first_of(delimiters, pos);
      tokenPos = str.find_first_not_of(delimiters, pos);
      
      if(std::string::npos != delimPos){
	if(std::string::npos != tokenPos){
	  if(tokenPos<delimPos){
	    tokens.push_back(str.substr(pos,delimPos-pos));
	  }else{
	    tokens.push_back("");
	  }
	}else{
	  tokens.push_back("");
	}
	pos = delimPos+1;
      } else {
	if(std::string::npos != tokenPos){
	  tokens.push_back(str.substr(pos));
	} else {
	  tokens.push_back("");
	}
	break;
      }
    }
    return tokens;
  }

  std::string MC_string_converter::zero_padding(const std::string& input,const int& zero_number)
  {
      std::string current_string=input;
      int current_length=input.size();
      for(int k=0;k<zero_number-current_length;k++)
          current_string=std::string("0")+current_string;
      return current_string;
  }
  std::string MC_string_converter::to_lower(const std::string& input)
  {
      std::string res("");
      for(unsigned int k=0;k<input.length();++k)
          res+=std::tolower(input[k],std::cout.getloc ());
      return res;
  }
  std::string MC_string_converter::to_upper(const std::string& input)
  {
      std::string res("");
      for(unsigned int k=0;k<input.length();++k)
          res+=std::toupper(input[k],std::cout.getloc ());
      return res;
  }

  std::vector <std::string> MC_string_converter::delete_empty(const std::vector <std::string>& in)
  {
      std::vector <std::string> res;

      int N=in.size();
      for(int k=0;k<N;++k)
          if(in[k]!=" " && in[k]!="")
              res.push_back(in[k]);
      return res;
  }

  MC_file_helper MC_file_helper::copy(const std::string& input_filename,const std::string& output_filename)
  {
      std::ifstream input(input_filename.c_str(),std::ifstream::in);
      std::ofstream output(output_filename.c_str(),std::ifstream::out);

      if(input.is_open()==false)
      {std::cout<<"Error in MC_file_helper::copy("<<input_filename<<","<<output_filename<<"), the file "<<input_filename<<" cannot be opened"<<std::endl;exit(-1);}
      if(output.is_open()==false)
      {std::cout<<"Error in MC_file_helper::copy("<<input_filename<<","<<output_filename<<"), the file "<<output_filename<<" cannot be opened"<<std::endl;exit(-1);}

      char buffer[static_cast<int>(pow(2,16))];
      while(input.good())
      {
          input.getline(buffer,static_cast<int>(pow(2,16)));
          if(input.good())
              output<<std::string(buffer)+"\n";
      }


      input.close();
      output.close();

      return MC_file_helper();
  }

}
