/*
**    TP ETI3 CPE Lyon
**    Copyright (C) 2009 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <MC_v3d.hpp>

#include <MC_trackball.hpp>



namespace mesh_conv
{

    MC_trackball::MC_trackball()
    {disc_radius=0.8;}

    MC_trackball::~MC_trackball(){}

    double MC_trackball::project_to_disc(const double& x,const double& y) const
    {
        double n=sqrt(x*x+y*y);
        if(n<disc_radius*0.707107)
            return sqrt(disc_radius*disc_radius-n*n);
        else
            return disc_radius*disc_radius/(2.0*n);
    }

    MC_trackball& MC_trackball::set_2d_coords(const double& x0,const double& y0,const double& x1,const double& y1)
    {
        //axis of the rotation
        MC_v3d axis;
        // angle of the rotation
        double phi;

        //security to not move if start=end
        double epsilon=0.0001;
        if( pow((x0-x1)*(x0-x1)+(y0-y1)*(y0-y1),0.5)<epsilon)
            d_q=MC_quaternion(1,0,0,0);
        else
        {
            // temp position
            MC_v3d p1,p2;


            //printf("(%f,%f),(%f,%f)\n",x0,y0,x1,y1);
            p1=MC_v3d(x0,y0,project_to_disc(x0,y0));
            p2=MC_v3d(x1,y1,project_to_disc(x1,y1));
            //std::cout<<p1<<","<<p2<<"  "<<p1-p2<<std::endl;

            axis = (p1.cross(p2)).normalized();

            MC_v3d u=p1-p2;
            double t=u.norm()/(2*disc_radius);
            t=fmin(fmax(t,-1.0),1.0); //clamp
            phi = 2.0*asin(t);

            //compute quaternion to apply
            d_q=MC_quaternion(axis,phi);
        }
        return *this;
    }

    MC_trackball& MC_trackball::apply_rotation()
    {current_q=d_q.conjugated()*current_q;return *this;}
    MC_trackball& MC_trackball::no_motion()
    {d_q=MC_quaternion(1,0,0,0);return *this;}


    const MC_quaternion& MC_trackball::quaternion() const{return current_q;}
    MC_quaternion& MC_trackball::quaternion(){return current_q;}

    const MC_quaternion& MC_trackball::d_quaternion() const {return d_q;}
    MC_quaternion& MC_trackball::d_quaternion() {return d_q;}

}
