/** TP ETI3 - CPE Lyon - Decembre 2009 */
/** Deformation de maillage */
/** version etudiante */


#ifndef _MESH_HPP_
#define _MESH_HPP_

#include <MC_v3d.hpp>

#include <iostream>
#include <vector>
#include <map>
#include <list>

namespace cpe_mesh
{

    /** \brief basic mesh class


    */
    class mesh
    {


    private:

        /** \brief vertex vector (size=3xN_vertex) */
        std::vector <double> vertex;
        /** \brief connectivity vector (only triangles) (size=3xN_poly)*/
        std::vector <int> connectivity;

    public:

        //*********************************************//
        //*********************************************//
        // CONSTRUCTORS
        //*********************************************//
        //*********************************************//

        /** \brief empty constructor */
        mesh();
        /** \brief constructor from std::pair <vertex,connectivity> */
        mesh(const std::pair <std::vector<double>,std::vector<int> >& input);


        //*********************************************//
        //*********************************************//
        // Get value
        //*********************************************//
        //*********************************************//

        /** \brief get vertex k (to do)*/
        mesh_conv::MC_v3d get_vertex(const unsigned int& k_vertex) const;
        /** \brief get connectivity polygon (to do)*/
        std::vector<int> get_connectivity(const unsigned int& k_polygon) const;

        /** \brief set a 3D vertex (to do)*/
        mesh& set_vertex(const unsigned int& k_vertex,const mesh_conv::MC_v3d& new_vertex);
        /** \brief set connectivity (to do)*/
        mesh& set_connectivity(const unsigned int& k_polygon,const std::vector<int>& new_triangle);


        //*********************************************//
        //*********************************************//
        // Size
        //*********************************************//
        //*********************************************//

        /** \brief number of vertices */
        unsigned int vertex_number() const;
        /** \brief number of triangles */
        unsigned int triangle_number() const;

        //*********************************************//
        //*********************************************//
        // Connectivity information
        //*********************************************//
        //*********************************************//

        /** \brief build one ring */
        std::map<int,std::list<int> > ring() const;
        /** \brief build one star */
        std::map<int,std::list<int> > star() const;



        //*********************************************//
        //*********************************************//
        // Normal
        //*********************************************//
        //*********************************************//

        /** \brief compute per vertex normal */
        std::vector <double> normal_vertex() const;
        /** \brief compute per polygon normal */
        std::vector <double> normal_polygon() const;

        //*********************************************//
        //*********************************************//
        // INTERNAL ACCESS
        //*********************************************//
        //*********************************************//

        /** \brief get internal vertex */
        const std::vector<double>& get_vertex() const;
        /** \brief get internal vertex */
        std::vector<double>& get_vertex();

        /** \brief get internal connectivity */
        const std::vector <int>& get_connectivity() const;
        /** \brief get internal connectivity */
        std::vector <int>& get_connectivity();

        //*********************************************//
        //*********************************************//
        // Assert
        //*********************************************//
        //*********************************************//

        /** \brief check vertex index are within the bounds */
        void assert_vertex(const unsigned int& k_vertex) const
        {if(3*k_vertex+2>=vertex.size()){std::cout<<"Error in mesh::assert_vertex("<<k_vertex<<"), vertex.size()="<<vertex.size()<<std::endl;exit(-1);}}

        /** \brief check polygon index are within the bounds */
        void assert_connectivity(const unsigned int& k_connectivity) const
        {if(3*k_connectivity+2>=connectivity.size()){std::cout<<"Error in mesh::assert_connectivity("<<k_connectivity<<"), connectivity.size()="<<connectivity.size()<<std::endl;exit(-1);}}


        //*********************************************//
        //*********************************************//
        // IN/OUT
        //*********************************************//
        //*********************************************//

        /** \brief load off file */
        static mesh load_off_file(const std::string& filename);
        /** \brief save mesh as off file */
        const mesh& save_off_file(const std::string& filename) const;


        //*********************************************//
        //*********************************************//
        // Deformation
        //*********************************************//
        //*********************************************//

        /** \brief scale and translate the mesh to fit into the unit cube */
        mesh& scale_unity();





    };
}

#endif
