/** TP ETI3 - CPE Lyon - Decembre 2009 */
/** Deformation de maillage */
/** version etudiante */

#include <MC_v3d.hpp>
#include <MC_navigator_tool.hpp>
#include <MC_double_vector.hpp>
#include <MC_matrix.hpp>
#include <MC_string_helper.hpp>

#include <mesh.hpp>

#include <stdlib.h>
#include <iostream>
#include <GL/glut.h>
#include <cmath>
#include <string>
#include <errno.h>
#include <string>
#include <cstdio>

#include <vector>

using namespace cpe_mesh;
using mesh_conv::MC_matrix ;
using mesh_conv::MC_v3d ;
using mesh_conv::MC_navigator_tool ;
using namespace std;

//glut functions
static void display_callback();
static void reshape_callback (int width, int height);
static void keyboard_callback(unsigned char key,int x,int y);
static void idle_callback(int value);
static void motion_callback (int x, int y);
static void mouse_callback (int button, int action, int x, int y);


//main functions
void draw_orientation();
void draw_pointer();
void draw_fps();
void draw_extra();
void draw_sphere(const MC_v3d& center,const double& radius);

int setup_opengl(int argc,char **argv);
int draw_scene(); //draw the model
int load_model();
double get_rand();//just a rand
int evolve();



//GLOBAL VARIABLES
//************************************//
MC_navigator_tool navigator;
int counter=0;
bool is_draw_grid=false;

cpe_mesh::mesh sphere;//saved sphere mesh
cpe_mesh::mesh mesh_0;//the original mesh
cpe_mesh::mesh mesh_tp;//the deformed mesh
std::map<int,std::list<int> > ring; //1-ring
std::vector <double> normal,color; //normal and color of the mesh


int main(int argc, char *argv[])
{
  cout<<"**hello**\n"<<endl;
  std::cout<<"run "<<argv[0]<<" with "<<argc-1<<" parameters ... \n"<<std::endl;


  //*************//
  //start OpenGL
  //*************//
  navigator.x_screen_size_1=800;
  navigator.y_screen_size_1=800;
  setup_opengl(argc,argv);

  //*************//
  //load sphere
  //*************//
  sphere=mesh::load_off_file("data/sphere.off");

  //*************//
  //load  models
  //*************//
  load_model();

  //*************//
  //run
  //*************//
  std::cout<<"**RUN**\n"<<std::endl;
  std::cout<<"Navigation with keys:\n\n"<<std::endl;
  std::cout<<"==========================\n"<<std::endl;
  std::cout<<"Space Displacement"<<std::endl;
  std::cout<<"  z    r"<<std::endl;
  std::cout<<"q s d  f\n"<<std::endl;
  std::cout<<"and mouse with drag & drop\n"<<std::endl;
  std::cout<<"W: wireframe mode"<<std::endl;
  std::cout<<"F: filled mode"<<std::endl;
  std::cout<<"G: Grid mode"<<std::endl;
  std::cout<<"X: Print cursor coordinates"<<std::endl;
  std::cout<<"Q: exit"<<std::endl;
  std::cout<<"==========================\n"<<std::endl;
  glutMainLoop ();


  return 0;
}















//load the model
int load_model()
{


  //*********************************//
  //LOAD MESH
  //*********************************//

  //std::string filename("data/dino_0.off");
  //std::string filename("data/dino_1.off");
  //std::string filename("data/camel.off");
  std::string filename("data/pig.off");
  mesh_tp=mesh::load_off_file(filename);

  //scale the mesh
  mesh_tp.scale_unity();

  //compute normals
  normal=mesh_tp.normal_vertex();

  //compute the colors
  color.resize(normal.size());
  for(int k=0,N=color.size();k<N;++k)
      color[k]=0.9*std::abs(normal[k]);

  //save original mesh
  mesh_0=mesh_tp;

  return 0;
}






int evolve()
{
    //set-up timer
    int ut=glutGet(GLUT_ELAPSED_TIME);
    double t;t=(ut-int(ut/2500)*2500)/2500.0;


    //************************//
    //******Deform mesh*******//
    //************************//




    //************************//
    //************************//

    return 0;
}


void draw_extra()
{
    //************************//
    //*********Draw***********//
    //************************//




    //************************//
    //************************//
}









//****************************************************************************//
//****************************************************************************//
//****************************************************************************//
//****************************************************************************//









//*******************//
//GLUT STUFF
//*******************//


//glut functions
int setup_opengl(int argc,char **argv)
{

    //init window
    glutInit (&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
    glutInitWindowPosition(0,0);
    glutInitWindowSize(navigator.x_screen_size_1,navigator.y_screen_size_1);
    glutCreateWindow("win_1");


    //OpenGL set
    glClearColor(1.0f,1.0f,1.0f,1.0f);
    glPointSize(5);
    glLineWidth(2);

    glShadeModel(GL_SMOOTH);
    glEnable(GL_COLOR_MATERIAL);
    //glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE );
    glEnable(GL_DEPTH_TEST);
    glDepthMask(GL_TRUE);


    glFrontFace(GL_CCW);//clockwise CW or CCW
    glDisable(GL_CULL_FACE);
    //glEnable(GL_CULL_FACE);

    glDisable(GL_BLEND);
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_FOG);

    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glFrontFace( GL_CW ) ;
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);

    glEnable(GL_LINE_SMOOTH);
    glHint(GL_LINE_SMOOTH_HINT,GL_NICEST);

    //callback
    glutDisplayFunc (display_callback);
    glutReshapeFunc (reshape_callback);
    glutKeyboardFunc(keyboard_callback);
    glutTimerFunc(33,idle_callback,1);
    glutMotionFunc (motion_callback);
    glutMouseFunc (mouse_callback);

    return 0;
}



//display
static void display_callback()
{


    //vieport (all screen)
    glViewport (0, 0, navigator.x_screen_size_1,navigator.y_screen_size_1);
    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    //Projection
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity ();
    mesh_conv::MC_double_vector proj_param=navigator.projection_param();
    glFrustum(proj_param[0], -proj_param[0], proj_param[1], -proj_param[1], proj_param[2], proj_param[3]);


    //Modelview
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glTranslated(0,0,static_cast<GLdouble>(navigator.dist_cam1)); //zoom
    MC_matrix temp_rotation = navigator.trackball_cam1.quaternion().conjugated().matrix().to_matrix4();
    glMultMatrixd(temp_rotation.pointer()); //rotation
    glTranslated(static_cast<GLdouble>(navigator.tr_cam1[0]),static_cast<GLdouble>(navigator.tr_cam1[1]),static_cast<GLdouble>(navigator.tr_cam1[2])); //translation


    //draw
    draw_scene();

    //draw_pointer
    draw_pointer();

    //draw orientation (clear DEPTH-BUFFER)
    draw_orientation();

    //update fps
    navigator.increase_frame_number();
    navigator.current_time=glutGet(GLUT_ELAPSED_TIME);
    if(navigator.delta_time()>1000)
    {navigator.update_fps();}

    //draw fps
    draw_fps();

    //error OpenGL_drawer_2
    GLenum errCode;
    const GLubyte *errString;
    if((errCode=glGetError()) != GL_NO_ERROR)
    {
        errString = gluErrorString(errCode);
        std::cout<<"OPENGL ERROR : "<<errString<<std::endl;
        exit(-2);
    }

    counter++;


  //redraw
  glFlush ();
  glutSwapBuffers ();

}


static void reshape_callback (int width, int height)
{
    navigator.x_screen_size_1=width;navigator.y_screen_size_1=height;
    glViewport(0,0, width, height);
}


//keyboard
static void keyboard_callback(unsigned char key,int x,int y)
{

  x=0;y=0;
  double dL=0.001*(1+100*abs(navigator.dist_cam1));

  #define GLUT_SPACE 32


  switch(key)
    {
    case 'Q'://quit
      printf("\n[EXIT OK]\n\n");
      exit(0);
      break;


    case 'z':
       navigator.go_forward_trackball_cam1(dL);
      break;
    case 's':
       navigator.go_forward_trackball_cam1(-dL);
      break;
    case 'q':
      navigator.go_right_trackball_cam1(-dL);
      break;
    case 'd':
      navigator.go_right_trackball_cam1(dL);
      break;
    case 'r':
      navigator.go_up_trackball_cam1(dL);
      break;
    case 'f':
      navigator.go_up_trackball_cam1(-dL);
      break;

    case 'G':
      is_draw_grid==1?is_draw_grid=0:is_draw_grid=1;
      break;

    case 'W'://wireframe
      glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
      break;
    case 'F'://Filled
      glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
      break;

  case 'X'://cursor position
      std::cout<<"cursor position: "<<-navigator.tr_cam1<<std::endl;
      break;
    }

  glutPostRedisplay ();
}


static void idle_callback(int value)
{
  value=1;//useless

  glutPostRedisplay ();
  glutTimerFunc( 20, idle_callback, 1);


  evolve();
}



static void motion_callback (int x, int y)
{
    double x_old=navigator.x_previous,y_old=navigator.y_previous;
    double w=navigator.x_screen_size_1,h=navigator.y_screen_size_1;

    if ( (navigator.left_button) )
    {
        //trackball
        navigator.trackball_cam1.set_2d_coords((2.0*x_old-w)/w,(2.0*y_old-h)/h,(2*x-w)/w,(2*y-h)/h);
        navigator.trackball_cam1.apply_rotation();
        navigator.x_previous=int(x);navigator.y_previous=int(y);

    }
    else if ( (navigator.right_button))
    {
        double u = y-y_old;
        double fu = u/500.0; //factor of magnification

        navigator.dist_cam1 += (fabs(navigator.dist_cam1)+1)*fu; navigator.dist_cam1 = min(navigator.dist_cam1,0.0);
        navigator.y_previous=int(y);

    }

  glutPostRedisplay ();
}
static void mouse_callback (int button, int action, int x, int y)
{
    navigator.x_previous=x;
    navigator.y_previous=y;

  if (button == GLUT_LEFT_BUTTON && action == GLUT_DOWN) {
    navigator.left_button=true;
  }
  if (button == GLUT_LEFT_BUTTON && action == GLUT_UP) 
    navigator.left_button=false;

  if (button == GLUT_RIGHT_BUTTON && action == GLUT_DOWN) {
    navigator.right_button=true;
  }
  if (button == GLUT_RIGHT_BUTTON && action == GLUT_UP) 
    navigator.right_button=false;

  glutPostRedisplay ();
}

//just a rand number \in [0,1]
double get_rand()
{
  int precision=50000;
  return double(rand()%precision)/double(precision);
}


void draw_orientation()
{
    glDisable(GL_LIGHTING);
    //rotation
    glClear(GL_DEPTH_BUFFER_BIT);
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity ();
    glOrtho(-1,1,1,-1,0.1,10.0);
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    glTranslated(0.78,0.78,-1);
    // get pointer
    MC_matrix M=navigator.trackball_cam1.quaternion().conjugated().matrix().to_matrix4();
    // rotation
    glMultMatrixd(M.pointer());

    double L=0.2;
    glBegin(GL_LINES);
    glColor3d(1,0,0);
    glVertex3d(0,0,0);
    glVertex3d(L,0,0);
    glColor3d(0,1,0);
    glVertex3d(0,0,0);
    glVertex3d(0,L,0);
    glColor3d(0,0,1);
    glVertex3d(0,0,0);
    glVertex3d(0,0,L);
    glEnd();
    //pop matrices
    glPopMatrix();
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);

}


void draw_pointer()
{

    glPushMatrix();
    glLoadIdentity();

    glTranslatef(0,0,GLfloat(navigator.dist_cam1)); //zoom
    MC_matrix temp_m=navigator.trackball_cam1.quaternion().conjugated().matrix().to_matrix4();
    glMultMatrixd(temp_m.pointer()); //rotation

    //local basis
    double L=0.01;

    glBegin(GL_LINES);
    glColor3d(1,0,0);
    glVertex3d(-L,0,0);
    glVertex3d(L,0,0);
    glColor3d(0,1,0);
    glVertex3d(0,-L,0);
    glVertex3d(0,L,0);
    glColor3d(0,0,1);
    glVertex3d(0,0,-L);
    glVertex3d(0,0,L);
    glEnd();

    glPopMatrix();

}


void draw_fps()
{
    glDisable(GL_DEPTH_TEST);

    int int_fps=int(navigator.current_fps);
    int deci_fps=int(10*(navigator.current_fps-int_fps));
    double truncated_fps=int_fps+double(deci_fps)/10.0;

    std::string fps=(mesh_conv::MC_string_converter::to_string(truncated_fps)+" fps");

    glPushMatrix();
    glLoadIdentity();
    glRasterPos3f(-0.5f,-0.5f,-1.4f);
    for (unsigned int k=0,N=fps.length();k<N;++k)
        glutBitmapCharacter(GLUT_BITMAP_TIMES_ROMAN_24,fps[k]);
    glPopMatrix();

    glEnable(GL_DEPTH_TEST);
}

void draw_sphere(const MC_v3d& center,const double& radius)
{

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();

    glTranslated(center[0],center[1],center[2]);
    glScaled(radius,radius,radius);


    double *p_vertex=&sphere.get_vertex()[0];
    int* p_connectivity=&sphere.get_connectivity()[0];

    glColor3d(0.1,0.1,0.1);
    glLineWidth(2.0);

    glBegin(GL_LINES);
    for(int k=0,N=sphere.triangle_number();k<N;++k)
    {

        double *v0=p_vertex+3*p_connectivity[3*k+0];
        double *v1=p_vertex+3*p_connectivity[3*k+1];
        double *v2=p_vertex+3*p_connectivity[3*k+2];

        glVertex3dv(v0);glVertex3dv(v1);
        glVertex3dv(v1);glVertex3dv(v2);
        glVertex3dv(v2);glVertex3dv(v0);
    }
    glEnd();

    glPopMatrix();
}


//Draw the scene
int draw_scene()
{

    glEnable(GL_POLYGON_OFFSET_FILL);
    glPolygonOffset(5,5);

    double* p_vertex=&mesh_tp.get_vertex()[0];
    double* p_normal=&normal[0];
    double* p_color=&color[0];
    int* p_connectivity=&mesh_tp.get_connectivity()[0];

    glEnable(GL_LIGHTING);
    glEnable(GL_BLEND);
    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_COLOR_MATERIAL);

    glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(3,GL_DOUBLE,0,p_vertex);

    glEnableClientState(GL_NORMAL_ARRAY);
    glNormalPointer(GL_DOUBLE,0,p_normal);

    glEnableClientState(GL_COLOR_ARRAY);
    glColorPointer(3,GL_DOUBLE,0,p_color);

    glDrawElements(GL_TRIANGLES,mesh_tp.get_connectivity().size(),GL_UNSIGNED_INT,p_connectivity);

    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_NORMAL_ARRAY);

    glDisable(GL_POLYGON_OFFSET_FILL);
    glDisable(GL_LIGHTING);

    if(is_draw_grid==true)
    {
        glColor3d(0.1,0.1,0.1);
        glLineWidth(2.0);

        glBegin(GL_LINES);
        for(int k=0,N=mesh_tp.triangle_number();k<N;++k)
        {
            double *v0=p_vertex+3*p_connectivity[3*k+0];
            double *v1=p_vertex+3*p_connectivity[3*k+1];
            double *v2=p_vertex+3*p_connectivity[3*k+2];

            glVertex3dv(v0);glVertex3dv(v1);
            glVertex3dv(v1);glVertex3dv(v2);
            glVertex3dv(v2);glVertex3dv(v0);
        }
        glEnd();
    }



    draw_extra();

  return 0;
}
