/** TP ETI3 - CPE Lyon - Decembre 2009 */
/** Deformation de maillage */
/** version etudiante */

#include <cstdlib>
#include <fstream>



#include <MC_string_helper.hpp>
#include <MC_io_off.hpp>
#include <mesh.hpp>


namespace cpe_mesh
{
    mesh::mesh(){}
    mesh::mesh(const std::pair <std::vector<double>,std::vector<int> >& input)
    {
        vertex=input.first;
        connectivity=input.second;
    }


    mesh_conv::MC_v3d mesh::get_vertex(const unsigned int& k_vertex) const
    {
        assert_vertex(k_vertex);

        std::cout<<"mesh::MC_v3d get_vertex to do !"<<std::endl;
        exit(-1);
    }
    std::vector<int> mesh::get_connectivity(const unsigned int& k_polygon) const
    {
        assert_connectivity(k_polygon);

        std::cout<<"mesh::get_connectivity to do !"<<std::endl;
        exit(-1);
    }

    mesh& mesh::set_vertex(const unsigned int& k_vertex,const mesh_conv::MC_v3d& new_vertex)
    {
        assert_vertex(k_vertex);

        std::cout<<"mesh::set_vertex to do !"<<std::endl;
        exit(-1);
    }
    mesh& mesh::set_connectivity(const unsigned int& k_polygon,const std::vector<int>& new_triangle)
    {
        assert_connectivity(k_polygon);
        if(new_triangle.size()<3)
        {std::cout<<"Error in mesh::set_connectivity, size of triangle is "<<new_triangle.size()<<std::endl;exit(-1);}

        std::cout<<"mesh::set_connectivity get_vertex to do !"<<std::endl;
        exit(-1);

    }


    const std::vector<double>& mesh::get_vertex() const
    {return vertex;}
    std::vector<double>& mesh::get_vertex()
    {return vertex;}

    const std::vector <int>& mesh::get_connectivity() const
    {return connectivity;}
    std::vector <int>& mesh::get_connectivity()
    {return connectivity;}


    mesh mesh::load_off_file(const std::string& filename)
    {
        return mesh(mesh_conv::MC_io_off::read_off_file(filename));
    }
    const mesh& mesh::save_off_file(const std::string& filename) const
    {
        mesh_conv::MC_io_off::write_off_file(filename,vertex,connectivity);
        return *this;
    }
    unsigned int mesh::vertex_number() const{return vertex.size()/3;}
    unsigned int mesh::triangle_number() const{return connectivity.size()/3;}

    mesh& mesh::scale_unity()
    {
        if(vertex_number()<1)
            return *this;

        mesh_conv::MC_v3d x_max(vertex[0],vertex[1],vertex[2]);
        mesh_conv::MC_v3d x_min=x_max;

        //get bounding box
        for(int k=0,N=vertex_number();k<N;++k)
        {
            mesh_conv::MC_v3d x=mesh_conv::MC_v3d(vertex[3*k],vertex[3*k+1],vertex[3*k+2]);
            for(int k_dim=0;k_dim<3;++k_dim)
            {
                if(x[k_dim]<x_min[k_dim])
                    x_min[k_dim]=x[k_dim];
                if(x[k_dim]>x_max[k_dim])
                    x_max[k_dim]=x[k_dim];
            }
        }

        //scale
        mesh_conv::MC_v3d x_size=x_max-x_min;
        double x_size_max=std::max(std::max(x_size[0],x_size[1]),x_size[2]);

        double epsilon=0.00001;
        if(std::abs(x_size_max)<epsilon)
        {
            std::cout<<"Scaling failed, mesh is too small in mesh::scale_unity"<<std::endl;
            return *this;
        }

        mesh_conv::MC_v3d x_middle=(x_max+x_min)/2.0;
        for(int k=0,N=vertex_number();k<N;++k)
        {
            for(int k_dim=0;k_dim<3;++k_dim)
                vertex[3*k+k_dim]=(vertex[3*k+k_dim]-x_middle[k_dim])/x_size_max;
        }

        return *this;
    }

    std::map<int,std::list<int> > mesh::ring() const
    {
        std::map<int,std::list<int> > r;
        for(int k=0,N=triangle_number();k<N;++k)
        {
            for(int k_dim=0;k_dim<3;++k_dim)
            {
                int u0=connectivity[3*k+k_dim];
                int u1=connectivity[3*k+(k_dim+1)%3];

                r[u0].push_back(u1);
                r[u1].push_back(u0);
            }
        }

        return r;
    }
    std::map<int,std::list<int> > mesh::star() const
    {
        std::map<int,std::list<int> > s;
        for(int k=0,N=triangle_number();k<N;++k)
            for(int k_dim=0;k_dim<3;++k_dim)
                s[connectivity[3*k+k_dim]].push_back(k);

        return s;
    }

    std::vector <double> mesh::normal_vertex() const
    {
        std::map<int,std::list<int> > s=star();
        std::vector <double> n(vertex.size());
        std::vector <double> n_poly=normal_polygon();

        for(int k=0,N=vertex_number();k<N;++k)
        {

            mesh_conv::MC_v3d temp_normal;
            for(std::list<int>::const_iterator it=s[k].begin(),it_end=s[k].end();it!=it_end;++it)
                temp_normal += mesh_conv::MC_v3d(n_poly[3*(*it)],n_poly[3*(*it)+1],n_poly[3*(*it)+2]);

            temp_normal=temp_normal.normalized();
            for(int k_dim=0;k_dim<3;++k_dim)
                n[3*k+k_dim]=temp_normal[k_dim];
        }

        return n;

    }
    std::vector <double> mesh::normal_polygon() const
    {
        std::vector <double> n(connectivity.size());
        for(int k=0,N=triangle_number();k<N;++k)
        {
            mesh_conv::MC_v3d x0=mesh_conv::MC_v3d(vertex[3*connectivity[3*k+0]],vertex[3*connectivity[3*k+0]+1],vertex[3*connectivity[3*k+0]+2]);
            mesh_conv::MC_v3d x1=mesh_conv::MC_v3d(vertex[3*connectivity[3*k+1]],vertex[3*connectivity[3*k+1]+1],vertex[3*connectivity[3*k+1]+2]);
            mesh_conv::MC_v3d x2=mesh_conv::MC_v3d(vertex[3*connectivity[3*k+2]],vertex[3*connectivity[3*k+2]+1],vertex[3*connectivity[3*k+2]+2]);

            mesh_conv::MC_v3d xn=((x1-x0).normalized().cross((x2-x0).normalized())).normalized();
            for(int k_dim=0;k_dim<3;++k_dim)
                n[3*k+k_dim]=xn[k_dim];
        }
        return n;
    }

}
